# type: ignore

import re

from pystencilssfg import SfgContext, SfgComposer
from pystencilssfg.configuration import SfgCodeStyle
from pystencilssfg.types import SrcType
from pystencilssfg.source_concepts import SrcObject
from pystencilssfg.source_components import (
    SfgClass,
    SfgMemberVariable,
    SfgConstructor,
    SfgMethod,
    SfgVisibility,
)
from pystencilssfg.emission import StringEmitter

from pystencils import fields, kernel

from util import normalize_whitespace as normalws

codestyle = SfgCodeStyle(run_clang_format=False)

def test_classes_explicit():
    ctx = SfgContext(codestyle=codestyle)
    sfg = SfgComposer(ctx)

    f, g = fields("f, g(1): double[2D]")

    @kernel
    def assignments():
        f[0, 0] @= 3 * g[0, 0]

    khandle = sfg.kernels.create(assignments)

    cls = SfgClass("MyClass")
    cls.default.append_member(SfgMethod("callKernel", sfg.call(khandle), const=True))

    cls.default.append_member(
        SfgMethod(
            "inlineConst",
            sfg.seq("return -1.0;"),
            return_type=SrcType("double"),
            inline=True,
            const=True,
        )
    )

    cls.default.append_member(
        SfgMethod(
            "awesomeMethod",
            sfg.seq("return 2.0f;"),
            return_type=SrcType("float"),
            inline=False,
            const=True,
        )
    )

    cls.default.append_member(SfgMemberVariable("stuff", "std::vector< int >"))

    cls.default.append_member(
        SfgConstructor([SrcObject("stuff", "std::vector< int > &")], ["stuff_(stuff)"])
    )

    ctx.add_class(cls)

    emitter = StringEmitter(ctx)
    header_code = normalws(emitter.get_header_code())

    structure_re = re.compile(
        r"class MyClass\s*\{\s*void callKernel\s*\(.*\)\s*const;\s*"
        r"double inlineConst\s*\(\)\s*const\s*\{\s*return -1\.0;\s*\}\s*"
        r"float awesomeMethod\s*\(\)\s*const;\s*"
        r"std::vector<\s*int\s*> stuff;\s*"
        r"MyClass\s*\(std::vector<\s*int\s*>\s*&\s*stuff\)\s*:\s*stuff_\(stuff\)\s*\{\s*\}\s*"
        r"\};"
    )
    assert structure_re.search(header_code) is not None
